from typing import Tuple, TypeAlias, Self
import os
import random
import threading
import time

import protocol



MAX_HP = 100

TILE_SIZE = 32
SPRITE_WIDTH = 128
WINDOW_WIDTH = 1056

RED_SPAWN_POINT = 0x01, 0, 8 * TILE_SIZE
BLUE_SPAWN_POINT = 0x01, WINDOW_WIDTH - SPRITE_WIDTH, 8 * TILE_SIZE

Addr: TypeAlias = Tuple[str, int]



class Player:

    id: int
    addr: Addr
    token: bytes
    team: protocol.Team
    name: str
    last_heartbeat: float

    facing: protocol.FacingDirection
    hp: int
    map_id: int
    x: float
    y: float
    is_afk: bool
    shielded: bool
    greenshard_hp: int
    last_attacker: str


    def __init__(self, id: int, addr: Addr, team: protocol.Team, name: str, token: bytes):

        self.id = id
        self.addr = addr
        self.team = team
        self.name = name
        self.token = token
        self.last_heartbeat = time.time()
        self.is_afk = False
        self.shielded = False
        self.greenshard_hp = 0
        self.last_attacker = ""

        self.revive()


    def revive(self) -> tuple[int]:

        with player_lock:

            if self.team == protocol.Team.Red:
                self.facing = protocol.FacingDirection.Right
                respawn_map_id, respawn_x, respawn_y = RED_SPAWN_POINT

            else:
                self.facing = protocol.FacingDirection.Left
                respawn_map_id, respawn_x, respawn_y = BLUE_SPAWN_POINT

            self.hp = MAX_HP
            self.is_on_ground = False

            self.map_id, self.x, self.y = respawn_map_id, respawn_x, respawn_y

            return respawn_map_id, respawn_x, respawn_y


    def check_token(self, incoming_token: bytes) -> bool:
        with player_lock:
            return self.token == incoming_token


    def receive_damage(self, attack_damage: int) -> None:
        with player_lock:
            self.hp = min(max(0, self.hp - attack_damage), MAX_HP)


    def heal(self, quantity: int) -> None:
        with player_lock:
            self.hp = min(max(0, self.hp + quantity), MAX_HP)


    @staticmethod
    def generate_id() -> int:
        return random.getrandbits(16)


    @staticmethod
    def generate_token() -> bytes:
        return os.urandom(16)


    @staticmethod
    def get_players_by_team(team: protocol.Team) -> list[Self]:
        with player_lock:
            return [player for player in player_db.values() if player.team == team]


    @staticmethod
    def get_player_by_token(token: bytes) -> Self | None:
        with player_lock:
            for player in player_db.values():
                if player.check_token(token):
                    return player
        return None



player_db: dict[int, Player] = {}
player_lock = threading.RLock()
