import base64
import os
import pytermgui as ptg

import hsmm
import mafiachat

PALETTE_COLOR = "#2D5426"
PALETTE_COLOR_LIGHT = "#55A649"
PALETTE_ALT = "#F67400"
PALETTE_ALT_LIGHT = "#FDBC4B"
PALETTE_ERROR = "#F71735"

DEFAULT_USER = "@mafiaDEV"
DEFAULT_PASSWORD = "OnlyRequiredToSignAndSendMessages"
DEFAULT_MAFIACHAT = "localhost"
DEFAULT_HSMM = "localhost"


class ButtonContainer(ptg.Container):
    def __init__(
        self,
        user,
        onclick,
    ) -> None:
        self.user = user
        self.button = ptg.Button(self.user["id"], lambda *_: onclick(self.user))
        self.button.set_char("delimiter", ["", ""])
        self.button.parent_align = ptg.HorizontalAlignment.LEFT
        self.button.styles.label = ""
        self.button.styles.highlight = ""
        self.button.styles._current = self.button.styles.label
        super().__init__(self.button, box="ROUNDED")

    def on_hover(self, _) -> bool:
        return self.button.on_hover()

    def on_release(self, _) -> bool:
        return self.button.on_release()

    def handle_mouse(
        self,
        event: ptg.MouseEvent,
    ) -> bool:
        return self.button.handle_mouse(event)

    def handle_key(
        self,
        key: str,
    ) -> bool:
        return self.button.handle_key(key)


class Client:
    def __init__(
        self,
        manager: ptg.manager,
    ):
        self.user_id = None
        self.password = None
        self.mafiachat = os.environ.get("MAFIACHAT_HOSTNAME", None)
        self.hsmm = os.environ.get("HSMM_HOSTNAME", None)

        self.server_users = []
        self.server_version = 0
        self.server_messages = []

        self.user = None
        self.selected_user = None

        self.manager = manager

        self.configure()
        self.set_layout(self.manager)

        header = ptg.Window(
            f"[primary bold {PALETTE_COLOR_LIGHT}]Shardy[/] - A MafiaChat client implementation",
            box="EMPTY",
        )
        self.manager.add(header, assign="header")

        footer = ptg.Window(
            ptg.Splitter(
                ptg.Button("Send >>", lambda *_: self.send()),
                ptg.Button("Refresh", lambda *_: self.refresh()),
                ptg.Button("Settings", lambda *_: self.settings()),
                ptg.Button("Quit", lambda *_: self.manager.stop()),
            ),
            box="EMPTY",
        )
        self.manager.add(footer, assign="footer")

        self.users = ptg.Container(box="EMPTY")
        users_window = ptg.Window(
            self.users,
            title="Users",
            vertical_align=ptg.VerticalAlignment.TOP,
        )
        self.manager.add(users_window, assign="users")

        self.messages_header = ptg.Container("", "", box="EMPTY")
        self.messages = ptg.Container(box="EMPTY")
        messages_window = ptg.Window(
            self.messages_header,
            ptg.Container(),
            "",
            self.messages,
            title="Messages",
            vertical_align=ptg.VerticalAlignment.TOP,
            overflow=ptg.Overflow.SCROLL,
        )
        self.manager.add(messages_window, assign="messages")

        self.input = ptg.InputField(prompt=">> ", multiline=False)
        input_window = ptg.Window(
            ptg.Container(self.input, box="EMPTY"),
            title="Input",
            vertical_align=ptg.VerticalAlignment.TOP,
        )
        manager.add(input_window, assign="input")

    def configure(self):
        ptg.Button.set_char("delimiter", ["\[ ", " ]"])
        ptg.Button.styles.label = f"@surface dim {PALETTE_ALT}"
        ptg.Button.styles.highlight = f"@surface+1 dim {PALETTE_ALT}"
        ptg.Splitter.set_char("separator", "")
        ptg.Window.styles.border__corner = PALETTE_COLOR_LIGHT
        ptg.Window.styles.border_blurred__corner_blurred = PALETTE_COLOR

    def set_layout(
        self,
        manager: ptg.WindowManager,
    ) -> None:
        layout = ptg.Layout()
        layout.add_slot("header", height=3)
        layout.add_break()
        layout.add_slot("users", width=0.2)
        layout.add_slot("messages")
        layout.add_break()
        layout.add_slot("input", height=3)
        layout.add_break()
        layout.add_slot("footer", height=3)
        manager.layout = layout

    def fetch_and_update(self):
        try:
            self.server_users = mafiachat.get_users()
            self.server_version, self.server_messages = mafiachat.get_messages()
        except mafiachat.MafiaChatConnectionException as err:
            self.users.set_widgets([])
            self.messages.set_widgets([])
            self.print_error(f"\[MafiaChat server] {str(err)}")
            return
        self.update()

    def update(self):
        def _select(u):
            self.selected_user = u
            self.update_messages()

        self.user = None
        found = False
        widgets = []
        for user in self.server_users:
            if self.user_id == user["id"]:
                self.user = user
            else:
                if self.selected_user is not None and (
                    self.selected_user["id"] == user["id"]
                ):
                    found = True
                widgets.append(ButtonContainer(user, _select))
        if self.user is None:
            self.messages_header.set_widgets(["", ""])
            self.messages.set_widgets([])
            self.print_error(f"User not found on server: {self.user_id}")
            return
        self.users.set_widgets(
            ["", ptg.Label(f"You're [{PALETTE_COLOR_LIGHT}]{self.user['id']}"), ""]
            + widgets
        )
        if found:
            self.update_messages()
        else:
            self.selected_user = None
            self.messages_header.set_widgets(
                [
                    ptg.Label("[italic]Please select a conversation on the left panel"),
                    "",
                ]
            )
            self.messages.set_widgets([])

    def update_messages(self):
        self.messages_header.set_widgets(
            [
                ptg.Label(f"[bold {PALETTE_ALT_LIGHT}]{self.selected_user['name']}"),
                ptg.Label(
                    f"[{PALETTE_ALT}]{self.selected_user['id']}[/] / [italic {PALETTE_ALT}]{self.selected_user['description']}"
                ),
            ]
        )
        widgets = []
        for message in self.server_messages:
            if (message["from"] == self.user["id"]) and (
                message["to"] == self.selected_user["id"]
            ):
                name = f"[{PALETTE_COLOR_LIGHT}]{self.user['name']}"
            elif (message["from"] == self.selected_user["id"]) and (
                message["to"] == self.user["id"]
            ):
                name = f"[{PALETTE_ALT_LIGHT}]{self.selected_user['name']}"
            else:
                continue
            widgets.append(
                ptg.Label(
                    f"{name}>[/] {message['content']}",
                    parent_align=ptg.HorizontalAlignment.LEFT,
                )
            )
        if len(widgets) == 0:
            widgets.append(
                ptg.Label(
                    "No messages",
                    parent_align=ptg.HorizontalAlignment.LEFT,
                )
            )
        self.messages.set_widgets(widgets)

    def print_error(
        self,
        msg: str,
    ):
        widgets = self.messages._widgets
        widgets.append(
            ptg.Label(
                f"[{PALETTE_ERROR}]*** {msg} ***",
                parent_align=ptg.HorizontalAlignment.LEFT,
            )
        )
        self.messages.set_widgets(widgets)

    def send(self):
        if self.selected_user is None:
            self.print_error("Select a conversation first")
            return
        message = self.input.value
        message_b64 = base64.b64encode(message.encode())
        self.input._lines = [""]
        try:
            data, sig = hsmm.sign(
                self.user_id,
                self.password,
                self.selected_user["id"],
                message_b64,
                1337,
                1337,
            )
        except (hsmm.HSMMConnectionException, hsmm.HSMMException) as err:
            self.print_error(f"\[HSMM server] {str(err)}")
            return
        try:
            self.server_version, self.server_messages = mafiachat.post_messages(
                data, sig
            )
        except mafiachat.MafiaChatConnectionException as err:
            self.print_error(f"\[MafiaChat server] {str(err)}")
            return
        self.update()

    def refresh(self):
        version = mafiachat.get_version()
        if version != self.server_version:
            self.fetch_and_update()

    def settings(
        self,
        init: bool = False,
    ) -> None:
        modal = None

        def _close(*_):
            self.manager.remove(modal)

        def _update(*_):
            self.password = field_password.value
            if (
                (self.mafiachat != field_mafiachat.value)
                or (self.hsmm != field_hsmm.value)
                or (self.user_id != field_user_id.value)
            ):
                self.mafiachat = field_mafiachat.value
                os.environ.update([("MAFIACHAT_HOSTNAME", self.mafiachat)])
                self.hsmm = field_hsmm.value
                os.environ.update([("HSMM_HOSTNAME", self.hsmm)])
                self.user_id = field_user_id.value
                self.fetch_and_update()
            _close()

        field_mafiachat = ptg.InputField(
            self.mafiachat if self.mafiachat is not None else DEFAULT_MAFIACHAT,
            prompt="MafiaChat: ",
        )
        field_hsmm = ptg.InputField(
            self.hsmm if self.hsmm is not None else DEFAULT_HSMM,
            prompt="HSMM: ",
        )

        field_user_id = ptg.InputField(
            self.user_id if self.user_id is not None else DEFAULT_USER,
            prompt="User: ",
        )
        field_password = ptg.InputField(
            self.password if self.password is not None else DEFAULT_PASSWORD,
            prompt="HSMM Password: ",
        )

        if init:
            button_splitter = ptg.Splitter(
                ptg.Button("Start MafiaChatting !", _update),
            )
        else:
            button_splitter = ptg.Splitter(
                ptg.Button("Apply", _update),
                ptg.Button("Cancel", _close),
            )

        modal = self.manager.alert(
            "",
            "Settings",
            "",
            ptg.Window(
                field_mafiachat,
                field_hsmm,
                title="Server hostnames",
                box="ROUNDED",
            ),
            "",
            ptg.Window(
                field_user_id,
                field_password,
                title="User",
                box="ROUNDED",
            ),
            "",
            ptg.Container(
                button_splitter,
                box="EMPTY",
            ),
            "",
            width=70,
        )


def main():
    with ptg.WindowManager() as manager:
        client = Client(manager)
        client.settings(True)


if __name__ == "__main__":
    main()
