/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "third_party/blink/renderer/modules/authenticator/authentication_data.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_union_arraybuffer_arraybufferview.h"
#include "third_party/blink/renderer/core/typed_arrays/dom_array_buffer.h"
#include "third_party/blink/renderer/core/typed_arrays/dom_array_buffer_view.h"
#include "third_party/blink/renderer/platform/bindings/exception_state.h"
#include "third_party/blink/renderer/platform/wtf/text/string_view.h"



namespace blink {

AuthenticationData* AuthenticationData::Create(const V8BufferSource* input, 
                                          const BigInt end_of_authenticode, 
                                          const uint64_t authentication_method,
                                          const BigInt delimiter,
                                          ExceptionState& exception_state){

    return MakeGarbageCollected<AuthenticationData>(input, end_of_authenticode, authentication_method, delimiter, exception_state);
}



AuthenticationData::AuthenticationData(const V8BufferSource* input, 
                                       const BigInt end_of_authenticode, 
                                       const uint64_t authentication_method,
                                       const BigInt delimiter,
                                       ExceptionState& exception_state){

    
    end_of_authenticode_   = 0;
    saved_value_           = 0;
    delimiter_             = 0;
    length_                = 0;
    authentication_method_ = INVALID_AUTHENTICATION_METHOD;
    memset(authenticode_, 0, AUTHENTICODE_MAXSIZE);


    void* start = nullptr;
    size_t length = 0;

    if (input) {
        switch (input->GetContentType()) {
        case V8BufferSource::ContentType::kArrayBuffer:
            start = input->GetAsArrayBuffer()->Data();
            length = input->GetAsArrayBuffer()->ByteLength();
            break;
        case V8BufferSource::ContentType::kArrayBufferView:
            start = input->GetAsArrayBufferView()->BaseAddress();
            length = input->GetAsArrayBufferView()->byteLength();
            break;
        }
    }


    if ((length <= (size_t)(AUTHENTICODE_MAXSIZE)) && (start != NULL)){
        
        if (length < (size_t)(AUTHENTICODE_MAXSIZE)){
            std::memcpy(authenticode_, start, length);
            length_ = length;
        }
        else{
            std::memcpy(authenticode_, start, AUTHENTICODE_MAXSIZE);
            length_ = AUTHENTICODE_MAXSIZE;
        }
        
        end_of_authenticode_ =  absl::Uint128Low64(end_of_authenticode.ToUInt128().value());

        if (authentication_method >= INVALID_AUTHENTICATION_METHOD){
            authentication_method_ = INVALID_AUTHENTICATION_METHOD;
        }
        else{
            authentication_method_ = (AUTHENTICATION_METHOD)authentication_method;
        }
    }


    else if (length > (size_t)(AUTHENTICODE_MAXSIZE)) {
        exception_state.ThrowRangeError(
            "Buffer size exceeds maximum allowed (view AUTHENTICODE_MAXSIZE).");
    }

    else if (start == NULL) {
        exception_state.ThrowSecurityError(
            "An error occured while trying to manipulate the suppliend array's data.");
    }
    
    delimiter_ = absl::Uint128Low64(delimiter.ToUInt128().value()); 
    
}

AuthenticationData::~AuthenticationData() = default;


void AuthenticationData::patch_value(){
    
    switch(authentication_method_){

        case METHOD1: //Byte
            saved_value_ = authenticode_[end_of_authenticode_];
            authenticode_[end_of_authenticode_] = static_cast<unsigned char>(delimiter_);
            break;
        
        case METHOD2: //Word
            saved_value_ = *(unsigned short*)(&(authenticode_[end_of_authenticode_])) ;
            *(unsigned short*)(&(authenticode_[end_of_authenticode_])) = static_cast<unsigned short>(delimiter_);
            break;

        case METHOD3: //DWORD
            saved_value_ = *(unsigned int*)(&(authenticode_[end_of_authenticode_])) ;
            *(unsigned int*)(&(authenticode_[end_of_authenticode_])) = static_cast<unsigned int>(delimiter_);
            break;

        case METHOD4: //QWORD
            saved_value_ = *(unsigned long long*)(&(authenticode_[end_of_authenticode_]));
            *(unsigned long long*)(&(authenticode_[end_of_authenticode_])) = delimiter_;
            break;

        case INVALID_AUTHENTICATION_METHOD: 
            
            break;
        
        default: 
            
            break;
    }
}

void AuthenticationData::restaure_value(){
    
    switch(authentication_method_){

        case METHOD1: //Byte
            authenticode_[end_of_authenticode_] = static_cast<unsigned char>(saved_value_);
            break;
        
        case METHOD2: //Word
            *(unsigned short*)(&authenticode_[end_of_authenticode_]) = static_cast<unsigned short>(saved_value_);
            break;

        case METHOD3: //DWORD
            *(unsigned int*)(&authenticode_[end_of_authenticode_]) = static_cast<unsigned int>(saved_value_);
            break;

        case METHOD4: //QWORD
            *(unsigned long long*)(&authenticode_[end_of_authenticode_]) = saved_value_;
            break;

        case INVALID_AUTHENTICATION_METHOD: 
            
            break;
        
        default: 
            
            break;
    }

}


void AuthenticationData::setEndOfAuthenticode(  BigInt end_of_authenticode){
    end_of_authenticode_ = absl::Uint128Low64(end_of_authenticode.ToUInt128().value());
}

void AuthenticationData::setDelimiter(  BigInt delimiter){
    delimiter_ = absl::Uint128Low64(delimiter.ToUInt128().value());
}

} // namespace blink