import socket
import struct
from cache import get_cached_client_decrypt, set_cached_client_decrypt
from config import *
from multiprocessing.pool import ThreadPool

def xor(a, b):
    return bytes([ca^cb for ca,cb in zip(a, b)])

def try_one(data):
    s = socket.create_connection((SERVER_HOST, SERVER_PORT))
    s.send(data)
    mtype = s.recv(1)
    payload = s.recv(1024)
    s.close()
    return data, mtype[0], payload

def try_position(known, ct_block, fillchar=0):
    datas = [b"\x01" + bytes([fillchar] * (16 - len(known) - 1) + [j] + known) + ct_block for j in range(256)]
    with ThreadPool(32) as p:
        good = []
        total = 0
        for data, mtype, payload in p.imap_unordered(try_one, datas):
            total += 1
            if (mtype, payload) != (0, b'Invalid ISO 7816-4 padding'):
                good.append(data[16 - len(known)])
            if len(good) == 1 and total > 1:
                return good[0]
            elif len(good) >= 2:
                return None
    return None

def _decrypt_block(ct_block):
    known = []
    fillchar = 0
    while len(known) < 16:
        print("...known:", bytes(known).hex())
        c = try_position(known, ct_block, fillchar=fillchar)
        if c is None:
            # oops, next bytes were flipped to 0x80 0x00 ... so all possibilities work - backtrack
            fillchar += 1
            known = known[1:]
            continue
        known = [0x80 ^ c] + known
    return bytes(known)

def decrypt_block(ct_block):
    pt = get_cached_client_decrypt(ct_block)
    if pt:
        return pt
    pt = _decrypt_block(ct_block)
    set_cached_client_decrypt(ct_block, pt)
    return pt

def decrypt_message(ct):
    pt = b""
    for i in range(16, len(ct), 16):
        iv = ct[i-16:i]
        ptb = decrypt_block(ct[i:i+16])
        ptb = xor(iv, ptb)
        pt += ptb
    return pt

def encrypt(data):
    assert len(data) % 16 == 0
    ct = b"\x00" * 16
    for i in reversed(range(0, len(data), 16)):
        ct = xor(data[i:i+16], decrypt_block(ct[:16])) + ct
    return ct

def fast_decrypt(ct):
    """ Fast decrypt using JSON error leak. Requires session ID + credential in config. Both session ID and credential should be hex."""
    from ast import literal_eval

    assert len(ct) % 16 == 0

    if all(get_cached_client_decrypt(ct[i:i+16]) for i in range(16, len(ct), 16)):
        return decrypt_message(ct)

    while 1:
        try:
            s = socket.create_connection((SERVER_HOST, SERVER_PORT))
            s.send(b"\x01" + encrypt(pad(bytes.fromhex(SESSION_ID) + struct.pack("<BH", 2, len(CREDENTIALS)) + CREDENTIALS.encode())))
            resp = s.recv(1024)
            if resp[0] == 0:
                raise Exception(resp[1:].decode())
            s.send(b"\x01" + encrypt(bytes.fromhex(SESSION_ID) + struct.pack("<BH", 7, 5 + 16 + len(ct)) + b"*****") + ct + encrypt(pad(b"")))
            resp = s.recv(1024)
            if not resp.startswith(b"\x00Unable to parse JSON: "):
                raise Exception("Bad response: " + repr(resp))
            break
        except Exception as e:
            print("oops", e)
            continue

    _, val = resp.split(b": ", 1)
    # drop ciphertext IV and junk from tail
    pt = literal_eval(val.decode())[21:-16]
    assert len(ct) == len(pt) + 16

    for i in range(16, len(ct), 16):
        iv = ct[i-16:i]
        set_cached_client_decrypt(ct[i:i+16], xor(iv, pt[i-16:i]))

    return pt

def pad(pt):
    from Crypto.Util.Padding import pad as _pad
    return _pad(pt, 16, "iso7816")

def unpad(pt):
    from Crypto.Util.Padding import unpad as _unpad
    return _unpad(pt, 16, "iso7816")
