from sha256_backdoored import K, message_schedule as sha256_message_schedule
from pyboolector import Boolector, BTOR_OPT_MODEL_GEN
import json

def _sigma0(btor, num: int):
    """As defined in the specification."""
    return (btor.Ror(num, 7) ^
           btor.Ror(num, 18) ^
           btor.Srl(num, 3))

def _sigma1(btor, num: int):
    """As defined in the specification."""
    return (btor.Ror(num, 17) ^
           btor.Ror(num, 19) ^
           btor.Srl(num, 10))

def _capsigma0(btor, num: int):
    """As defined in the specification."""
    return (btor.Ror(num, 2) ^
           btor.Ror(num, 13) ^
           btor.Ror(num, 22))

def _capsigma1(btor, num: int):
    """As defined in the specification."""
    return (btor.Ror(num, 6) ^
           btor.Ror(num, 11) ^
           btor.Ror(num, 25))

def _ch(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (~x & z)

def _maj(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (x & z) ^ (y & z)

def message_schedule(btor, msg_arr):
    assert len(msg_arr) == 16
    result = msg_arr[:]
    for t in range(16, 64):
        term1 = _sigma1(btor, result[t-2])
        term2 = result[t-7]
        term3 = _sigma0(btor, result[t-15])
        term4 = result[t-16]
        result.append(term1 + term2 + term3 + term4)
    return result

def step(btor, arr, msgs, index):
    a, b, c, d, e, f, g, h = arr
    t1 = h + _capsigma1(btor, e) + _ch(e, f, g) + K[index] + msgs[index]
    t2 = _capsigma0(btor, a) + _maj(a, b, c)

    h = g
    g = f
    f = e
    e = d + t1
    d = c
    c = b
    b = a
    a = t1 + t2
    return [a, b, c, d, e, f, g, h]

def add(arr1, arr2):
    return [a+b for a,b in zip(arr1, arr2)]

class FaultInjectionSolver:
    def __init__(self, baseline, errs):
        # errs: [(round, hash)]
        self.btor = Boolector()
        self.btor.Set_opt(BTOR_OPT_MODEL_GEN, 1)
        BitVec32 = self.btor.BitVecSort(32)
        UInt32 = lambda name: self.btor.Var(BitVec32, name)
        self.UInt32 = UInt32

        self.ihv = [UInt32(f"ihv{i}") for i in range(8)]
        self.mhv = [UInt32(f"mhv{i}") for i in range(8)]
        min_round = min(r for r, eh in errs)
        self.msg = [None] * min_round + [UInt32(f"msg{i}") for i in range(min_round, 64)]

        # baseline constraint
        arr = self.mhv[:]
        self.mhvs = {}
        for t in range(min_round, 64):
            self.mhvs[t] = arr
            arr = step(self.btor, arr, self.msg, t)
        arr = add(self.ihv, arr)
        self._add_hash_constraint(arr, baseline)

        # error constraints
        self.errs = []
        for i, (bp, eh) in enumerate(errs):
            ev = UInt32(f"err{i}")
            self.errs.append(ev)
            arr = self.mhvs[bp][:]
            arr[2] = arr[2] + ev
            for t in range(bp, 64):
                arr = step(self.btor, arr, self.msg, t)
            arr = add(self.ihv, arr)
            self._add_hash_constraint(arr, eh)

    def solve(self):
        res = self.btor.Sat()
        if res != self.btor.SAT:
            raise Exception("failed to solve")

        ihv = [int(k.assignment, 2) for k in self.ihv]
        msg = [k and int(k.assignment, 2) for k in self.msg]
        return (ihv, msg)

    def _add_hash_constraint(self, arr, hash):
        for i in range(8):
            hi = int.from_bytes(hash[i*4:i*4 + 4], "big")
            self.btor.Assert(arr[i] == hi)

data = json.load(open("collect-remote.json"))
baseline = bytes.fromhex(data["baseline"])

## Round 1: solve for an approximate ihv
errs = []
for bp in [60, 62]:
    for i in range(8):
        errs.append((bp, bytes.fromhex(data["03_%02d" % bp][i])))

solver = FaultInjectionSolver(baseline, errs)
ihv_arr, msg_arr = solver.solve()
print(ihv_arr, msg_arr)

# The result is accurate except that ihv[0] and ihv[4] are offset by a constant

## Round 2: solve for the message schedule (faster now that 7/8 of the ihv is known)
errs = []
for bp in range(40, 64, 2):
    for i in range(4):
        errs.append((bp, bytes.fromhex(data["03_%02d" % bp][i])))

solver = FaultInjectionSolver(baseline, errs)
for i in range(8):
    # The constant error is absorbed into M[63], so we can just use our current IHV guess as-is
    solver.btor.Assert(solver.ihv[i] == ihv_arr[i])

ihv_arr, msg_arr = solver.solve()
print(ihv_arr, msg_arr)

## Round 3: solve for the message itself (c2)
btor = Boolector()
btor.Set_opt(BTOR_OPT_MODEL_GEN, 1)
BitVec32 = btor.BitVecSort(32)
UInt32 = lambda name: btor.Var(BitVec32, name)

msg = message_schedule(btor, [UInt32(f"msg{i}") for i in range(16)])
for i in range(44, 62):
    btor.Assert(msg[i] == msg_arr[i])

res = btor.Sat()
assert res == btor.SAT

real_msg = b"".join([int(k.assignment, 2).to_bytes(4, "big") for k in msg[:16]])
print("Recovered real message (c2):", real_msg.hex())
real_M = sha256_message_schedule(real_msg)

## Round 4: solve for the precise ihv
errs = []
for bp in [60, 62]:
    for i in range(8):
        errs.append((bp, bytes.fromhex(data["03_%02d" % bp][i])))

solver = FaultInjectionSolver(baseline, errs)

for i in range(60, 64):
    solver.btor.Assert(solver.msg[i] == real_M[i])

ihv_arr, msg_arr = solver.solve()
print(ihv_arr, msg_arr)

real_ihv = b"".join([k.to_bytes(4, "big") for k in ihv_arr])
print("Recovered real oh-ihv (c1):", real_ihv.hex())

# Sanity check
from sha256_backdoored import sha256_hash_block
assert sha256_hash_block(real_ihv, real_msg) == baseline
