#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# Adapted from https://gist.githubusercontent.com/Pagliacii/774ed5d3ea78a36cdb0754be6a25408d/raw/a1bda62a53742599e140ef3ab499a5519863c0fa/win_api_test.py

import ctypes
import os
from ctypes import windll, wintypes, get_last_error, WinError, POINTER
from typing import Optional, NamedTuple
import struct

NULL = 0

GENERIC_ALL = 28
GENERIC_EXECUTE = 29
GENERIC_WRITE = 30
GENERIC_READ = 31

FILE_SHARE_DISABLE = 0x00000000
FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002
FILE_SHARE_DELETE = 0x00000004

CREATE_NEW = 1
CREATE_ALWAYS = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
TRUNCATE_EXISTING = 5

FILE_ATTRIBUTE_READONLY = 0x1
FILE_ATTRIBUTE_HIDDEN = 0x2
FILE_ATTRIBUTE_SYSTEM = 0x4
FILE_ATTRIBUTE_DIRECTORY = 0x10
FILE_ATTRIBUTE_ARCHIVE = 0x20
FILE_ATTRIBUTE_DEVICE = 0x40
FILE_ATTRIBUTE_NORMAL = 0x80
FILE_ATTRIBUTE_TEMPORARY = 0x100
FILE_ATTRIBUTE_SPARSE_FILE = 0x200
FILE_ATTRIBUTE_REPARSE_POINT = 0x400
FILE_ATTRIBUTE_COMPRESSED = 0x800
FILE_ATTRIBUTE_OFFLINE = 0x1000
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x2000
FILE_ATTRIBUTE_ENCRYPTED = 0x4000
FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x8000
FILE_ATTRIBUTE_VIRTUAL = 0x10000
FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x20000
FILE_ATTRIBUTE_RECALL_ON_OPEN = 0x40000
FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS = 0x400000

FILE_FLAG_BACKUP_SEMANTICS = 0x02000000
FILE_FLAG_DELETE_ON_CLOSE = 0x04000000
FILE_FLAG_NO_BUFFERING = 0x20000000
FILE_FLAG_OPEN_NO_RECALL = 0x00100000
FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000
FILE_FLAG_OVERLAPPED = 0x40000000
FILE_FLAG_POSIX_SEMANTICS = 0x0100000
FILE_FLAG_RANDOM_ACCESS = 0x10000000
FILE_FLAG_SESSION_AWARE = 0x00800000
FILE_FLAG_SEQUENTIAL_SCAN = 0x08000000
FILE_FLAG_WRITE_THROUGH = 0x80000000

INVALID_HANDLE_VALUE = 0xffff_ffff_ffff_ffff

PROCESS_QUERY_LIMITED_INFORMATION = 0x1000
TOKEN_QUERY = 0x8

TokenOwner = 4
SystemHandleInformation = 0x10
ObjectTypesInformation = 3

# ref: https://docs.microsoft.com/en-us/previous-versions/windows/desktop/legacy/aa379560(v=vs.85)
class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [
        ("nLength", wintypes.DWORD),
        ("lpSecurityDescriptor", wintypes.LPVOID),
        ("bInheritHandle", wintypes.BOOL)
    ]


# ref: https://docs.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-large_integer-r1
class LARGE_INTEGER(ctypes.Structure):
    _fields_ = [
        ("QuadPart", ctypes.c_longlong)
    ]

# ref: https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/ex/sysinfo/handle_table_entry.htm?ts=0,136
class SYSTEM_HANDLE_TABLE_ENTRY_INFO(ctypes.Structure):
    _fields_ = [
        ("UniqueProcessId", ctypes.c_ushort),
        ("CreatorBackTraceIndex", ctypes.c_ushort),
        ("ObjectTypeIndex", ctypes.c_byte),
        ("HandleAttributes", ctypes.c_byte),
        ("HandleValue", ctypes.c_short),
        ("Object", ctypes.c_ulonglong),
        ("GrantedAccess", ctypes.c_ulonglong),
    ]
    
    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(f"{fname}={getattr(self, fname):#x}" for fname, _ in self._fields_) + ")"

class OBJECT_TYPE_INFORMATION(NamedTuple):
    TypeName: str
    TotalNumberOfObjects: int
    TotalNumberOfHandles: int
    TotalPagedPoolUsage: int
    TotalNonPagedPoolUsage: int
    TotalNamePoolUsage: int
    TotalHandleTableUsage: int
    HighWaterNumberOfObjects: int
    HighWaterNumberOfHandles: int
    HighWaterPagedPoolUsage: int
    HighWaterNonPagedPoolUsage: int
    HighWaterNamePoolUsage: int
    HighWaterHandleTableUsage: int
    InvalidAttributes: int
    GenericMapping_Read: int
    GenericMapping_Write: int
    GenericMapping_Execute: int
    GenericMapping_All: int
    ValidAccessMask: int
    SecurityRequired: int
    MaintainHandleCount: int
    TypeIndex: int
    PoolType: int
    DefaultPagedPoolCharge: int
    DefaultNonPagedPoolCharge: int

def _filehandle_errcheck(result, func, args):
    if result == INVALID_HANDLE_VALUE:
        raise WinError()
    return result

def _handle_errcheck(result, func, args):
    if result is None:
        raise WinError()
    return result

def _bool_errcheck(result, func, args):
    if not result:
        raise WinError()


_CreateFileW = windll.kernel32.CreateFileW
_CreateFileW.argtypes = (
    wintypes.LPCWSTR,  # LPCWSTR               lpFileName
    wintypes.DWORD,    # DWORD                 dwDesiredAccess
    wintypes.DWORD,    # DWORD                 dwShareMode
    wintypes.LPVOID,   # LPSECURITY_ATTRIBUTES lpSecurityAttributes
    wintypes.DWORD,    # DWORD                 dwCreationDisposition
    wintypes.DWORD,    # DWORD                 dwFlagsAndAttributes
    wintypes.HANDLE,   # HANDLE                hTemplateFile
)
_CreateFileW.restype = wintypes.HANDLE
_CreateFileW.errcheck = _filehandle_errcheck

LPDWORD = POINTER(wintypes.DWORD)
LPOVERLAPPED = wintypes.LPVOID

_DeviceIoControl = windll.kernel32.DeviceIoControl
_DeviceIoControl.argtypes = [
        wintypes.HANDLE,                    # _In_          HANDLE hDevice
        wintypes.DWORD,                     # _In_          DWORD dwIoControlCode
        wintypes.LPVOID,                    # _In_opt_      LPVOID lpInBuffer
        wintypes.DWORD,                     # _In_          DWORD nInBufferSize
        wintypes.LPVOID,                    # _Out_opt_     LPVOID lpOutBuffer
        wintypes.DWORD,                     # _In_          DWORD nOutBufferSize
        LPDWORD,                            # _Out_opt_     LPDWORD lpBytesReturned
        LPOVERLAPPED]                       # _Inout_opt_   LPOVERLAPPED lpOverlapped
_DeviceIoControl.restype = wintypes.BOOL
_DeviceIoControl.errcheck = _bool_errcheck

_CloseHandle = windll.kernel32.CloseHandle
_CloseHandle.argtypes = (wintypes.HANDLE,)
_CloseHandle.restype = wintypes.BOOL
_CloseHandle.errcheck = _bool_errcheck

_CancelIoEx = windll.kernel32.CancelIoEx
_CancelIoEx.argtypes = [
        wintypes.HANDLE,                    # _In_          HANDLE hFile
        LPOVERLAPPED]                       # _Inout_opt_   LPOVERLAPPED lpOverlapped
_CancelIoEx.restype = wintypes.BOOL
_CancelIoEx.errcheck = _bool_errcheck

HANDLER_ROUTINE = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.DWORD)

_SetConsoleCtrlHandler = windll.kernel32.SetConsoleCtrlHandler
_SetConsoleCtrlHandler.argtypes = [
        HANDLER_ROUTINE,
        wintypes.BOOL]
_SetConsoleCtrlHandler.restype = wintypes.BOOL
_SetConsoleCtrlHandler.errcheck = _bool_errcheck

_NtQuerySystemInformation = windll.ntdll.NtQuerySystemInformation
_NtQuerySystemInformation.argtypes = [
        ctypes.c_int,
        wintypes.LPVOID,
        wintypes.ULONG,
        wintypes.PULONG]
_NtQuerySystemInformation.restype = ctypes.c_int
# _NtQuerySystemInformation.errcheck = _ntstatus_errcheck

_NtQueryObject = windll.ntdll.NtQueryObject
_NtQueryObject.argtypes = [
        wintypes.HANDLE,
        ctypes.c_int,
        wintypes.LPVOID,
        wintypes.ULONG,
        wintypes.PULONG]
_NtQueryObject.restype = ctypes.c_int
# _NtQueryObject.errcheck = _ntstatus_errcheck

_OpenProcess = windll.kernel32.OpenProcess
_OpenProcess.argtypes = [
    wintypes.DWORD,
    wintypes.BOOL,
    wintypes.DWORD,
]
_OpenProcess.restype = wintypes.HANDLE
_OpenProcess.errcheck = _handle_errcheck

_OpenProcessToken = windll.advapi32.OpenProcessToken
_OpenProcessToken.argtypes = [
    wintypes.HANDLE,
    wintypes.DWORD,
    POINTER(wintypes.HANDLE),
]
_OpenProcessToken.restype = wintypes.BOOL
_OpenProcessToken.errcheck = _bool_errcheck

_GetTokenInformation = windll.advapi32.GetTokenInformation
_GetTokenInformation.argtypes = [
        wintypes.HANDLE,
        ctypes.c_int,
        wintypes.LPVOID,
        wintypes.ULONG,
        wintypes.PULONG]
_GetTokenInformation.restype = wintypes.BOOL
_GetTokenInformation.errcheck = _bool_errcheck

# ugly hack to get ctypes base class
_CData = ctypes.c_int.__mro__[-2]

# ref: https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilew
def CreateFile(
    path: str,
    access: int=GENERIC_READ,
    share_mode: int=FILE_SHARE_DISABLE,
    security_attrs: Optional[SECURITY_ATTRIBUTES]=None,
    creation: int=OPEN_EXISTING,
    flags: int=FILE_ATTRIBUTE_NORMAL,
    template: wintypes.HANDLE=NULL,
) -> wintypes.HANDLE:

    # these two arguments below have not been testing.
    lpSecurityAttributes = ctypes.byref(security_attrs) if security_attrs else NULL

    return _CreateFileW(
        path, access, share_mode, lpSecurityAttributes,
        creation, flags, template
    )


def DeviceIoControl(
    handle: int,
    control_code: int,
    inbuf: None | bytes | bytearray | _CData=None,
    out: None | bytes | bytearray | _CData=None,
) -> int | bytes:
    if inbuf is None:
        _inbuf = NULL
        _insz = 0
    elif isinstance(inbuf, (bytes, bytearray)):
        _inbuf = inbuf
        _insz = len(inbuf)
    else:
        _inbuf = ctypes.byref(inbuf)
        _insz = ctypes.sizeof(inbuf)

    if out is None:
        _outbuf = NULL
        _outsz = 0
    elif isinstance(out, (bytes, bytearray)):
        _outbuf = ctypes.create_string_buffer(bytes(out), len(out))
        _outsz = len(out)
    else:
        _outbuf = ctypes.byref(out)
        _outsz = ctypes.sizeof(out)

    bytes_returned = wintypes.DWORD(0)
    _DeviceIoControl(handle, control_code, _inbuf, _insz, _outbuf, _outsz, ctypes.byref(bytes_returned), None)

    if isinstance(out, (bytes, bytearray)):
        return _outbuf[:bytes_returned.value]
    else:
        return bytes_returned.value


# ref: https://docs.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-closehandle
def CloseHandle(handle: wintypes.HANDLE) -> bool:
    return _CloseHandle(handle)

def CancelIoEx(handle: int):
    return _CancelIoEx(handle, None)

_console_ctrl_handlers = {}
def SetConsoleCtrlHandler(handler, add: bool):
    if add:
        if handler in _console_ctrl_handlers:
            raise Exception("already registered")
        wrapper = HANDLER_ROUTINE(handler)
        _SetConsoleCtrlHandler(wrapper, 1)
        _console_ctrl_handlers[handler] = wrapper
    else:
        wrapper = _console_ctrl_handlers[handler]
        _SetConsoleCtrlHandler(wrapper, 0)
        del _console_ctrl_handlers[handler]

def NtQuerySystemInformation(info_class: int) -> bytes:
    bufsize = 256
    retsize = ctypes.wintypes.ULONG(0)
    while 1:
        buf = ctypes.create_string_buffer(bufsize)
        status = _NtQuerySystemInformation(info_class, buf, bufsize, ctypes.byref(retsize))
        if status >= 0:
            return buf[:retsize.value]
        status &= 0xffff_ffff
        if status == 0xC0000004:
            # STATUS_INFO_LENGTH_MISMATCH
            bufsize = retsize.value
            continue
        raise OSError(0, "NtQuerySystemInformation failed: %08x" % status, "", status - (1<<32))

def NtQuerySystemInformation_SystemHandleInformation():
    res = NtQuerySystemInformation(SystemHandleInformation)
    count, = struct.unpack_from("<Q", res, 0)
    return (SYSTEM_HANDLE_TABLE_ENTRY_INFO * count).from_buffer_copy(res, 8)

def NtQueryObject(info_class: int, handle: int | None=None) -> bytes:
    bufsize = 256
    retsize = ctypes.wintypes.ULONG(0)
    while 1:
        buf = ctypes.create_string_buffer(bufsize)
        status = _NtQueryObject(handle, info_class, buf, bufsize, ctypes.byref(retsize))
        if status >= 0:
            return buf[:retsize.value]
        status &= 0xffff_ffff
        if status == 0xC0000004:
            # STATUS_INFO_LENGTH_MISMATCH
            bufsize = retsize.value
            continue
        raise OSError(0, "NtQueryObject failed: %08x" % status, "", status - (1<<32))

def NtQueryObject_ObjectTypesInformation():
    buf = NtQueryObject(ObjectTypesInformation)

    fmt = "<HH4xQ" "18IBBBx3I"
    sz = struct.calcsize(fmt)
    assert sz == 0x68

    count, = struct.unpack_from("<Q", buf, 0)
    res = []
    pos = 8
    for i in range(count):
        name_len, name_maxlen, name_buf, *info = struct.unpack_from(fmt, buf, pos)
        pos += sz
        name = buf[pos:pos + name_len].decode("utf-16-le")
        res.append(OBJECT_TYPE_INFORMATION(name, *info))
        pos += name_len + 2
        # pad
        pos += (-pos) % 8

    return res

def OpenProcess(access: int, inherit: bool, pid: int) -> int:
    return _OpenProcess(access, inherit, pid)

def OpenProcessToken(proc_handle: int, access: int) -> int:
    res = wintypes.HANDLE()
    _OpenProcessToken(proc_handle, access, ctypes.byref(res))
    return res.value

def GetTokenInformation(info_class: int, handle: int | None=None) -> bytes:
    bufsize = 256
    retsize = ctypes.wintypes.ULONG(0)
    while 1:
        buf = ctypes.create_string_buffer(bufsize)
        try:
            status = _GetTokenInformation(handle, info_class, buf, bufsize, ctypes.byref(retsize))
            return buf[:retsize.value]
        except Exception as e:
            print(e)
            break

def get_processes_and_names():
    # SystemFullProcessInformation
    buf = NtQuerySystemInformation(5)
    res = []
    pos = 0
    while pos < len(buf):
        next_offset, = struct.unpack_from("<I", buf, pos)
        thread_count, = struct.unpack_from("<I", buf, pos + 4)
        image_name_length, = struct.unpack_from("<H", buf, pos + 0x38)
        pid, ppid = struct.unpack_from("<QQ", buf, pos + 0x50)
        image_name_offset = pos + 0x100 + 0x50 * thread_count + 0x170 # xxx what's 0x170
        image_name = buf[image_name_offset:image_name_offset + image_name_length]
        res.append((pid, ppid, image_name.decode("utf-16-le")))
        if not next_offset:
            break
        pos += next_offset

    return res
